/*
 * Copyright 2002, 2014 IBM Corp.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.  
 *
 *  Author: Sami Salkosuo, sami.salkosuo@fi.ibm.com
*/
package clenotes.utils

import clenotes.Command
import clenotes.CommandLineArguments
import clenotes.Configuration
import clenotes.Logger
import java.io.BufferedReader
import java.io.File
import java.io.FileReader
import java.util.List
import java.util.Map
import java.util.Vector
import lotus.domino.Database
import lotus.domino.DateTime
import lotus.domino.DbDirectory
import lotus.domino.Document
import lotus.domino.DocumentCollection
import lotus.domino.EmbeddedObject
import lotus.domino.Item
import lotus.domino.MIMEEntity
import lotus.domino.MIMEHeader
import lotus.domino.RichTextItem
import lotus.domino.Session
import lotus.domino.Stream

class DominoUtils {

	def static Database getMailDatabase(Session notesSession) {

		/*Return users mail database. Arguments are session that is used to
  get mail database and opts. If options has replica-id option, it is used
  to get local mail db replica*/
		var DbDirectory dbDirectory = null
		var serverName = CommandLineArguments::getValue("server-name")
		dbDirectory = notesSession.getDbDirectory(serverName) //serverName may be null, opens default directory

		var replicaId = CommandLineArguments::getValue("replica-id")
		var databaseName = CommandLineArguments::getValue("database-name")

		return openDatabase(dbDirectory, databaseName, replicaId)

	}

	def static Database openDatabase(DbDirectory dbDirectory, String databaseName, String _replicaId) {

		var replicaId = _replicaId
		if (databaseName == null && replicaId == null) {

			if (CommandLineArguments::hasOption("local")) {
				var Database db = dbDirectory.openMailDatabase();
				replicaId = db.replicaID
				db.recycle
			} else {
				return dbDirectory.openMailDatabase()

			}
		}

		if (replicaId != null) {
			return dbDirectory.openDatabaseByReplicaID(replicaId)

		}

		if (databaseName != null) {
			return dbDirectory.openDatabase(databaseName, false)

		}

	}

	def static getItemValue(Document doc, String itemName) {

		//By default Document.getItemValue() returns Vector even if item type is not
		//list. This function returns value of element instead of Vector that holds value.
		var value = doc.getItemValue(itemName) as Vector<?>
		var Object rValue = null
		if (value != null && value.size > 0) {
			rValue = value.firstElement
		} else {
			if (itemName == "DeliveredDate") {

				//if DeliveredDate missing, check PostedDate
				value = doc.getItemValue("PostedDate") as Vector<?>
				if (value != null && value.size > 0) {
					rValue = value.firstElement
				}
			}
		}
		rValue
	}

	def static getItemValues(Document doc, String itemName) {
		var value = doc.getItemValue(itemName) as Vector<?>
		var str = new StringBuffer
		for (v : value) {
			str.append(v)
			str.append(', ')

		}
		var values = str.toString
		if (values.length >= 2) {
			values = values.substring(0, values.length - 2)
		}
		return values

	}

	def static sortDocumentCollectionIfNecessary(DocumentCollection docCollection, Database database, Command cmd) {
		var sortOrder = cmd.getOptionValue("sortorder")
		var sortField = cmd.getOptionValue("sortfield")

		//sortOrder='ASC',sortField='DATE'
		//sort-field!: Sort mails by field. Use values DATE, SUBJECT or FROM. Default is DATE.
		var allowedOrders = #['ASC', 'DESC']
		var allowedFields = #['DATE', 'SUBJECT', 'FROM']

		if (sortOrder != null || sortField != null) {

			//do sorting if either field is not empty
			if(sortOrder == null) sortOrder = allowedOrders.get(0)
			if(sortField == null) sortField = allowedFields.get(0)

			if (!allowedOrders.contains(sortOrder)) {

				throw new IllegalArgumentException(
					'Sort order not recognized. Allowed sort orders: ' + allowedOrders.join(","))
			}

			if (!allowedFields.contains(sortField)) {

				throw new IllegalArgumentException(
					'Sort field not recognized. Allowed sort orders: ' + allowedFields.join(","))
			}

			var newDocCollection = database.getProfileDocCollection("PROFILE_THAT_SHOULD_STILL_NOT_EXIST")

			Logger::log("Sort by " + sortField)
			var docsToBeSorted = newLinkedHashMap
			var n = 1
			var doc = docCollection.getNthDocument(n)
			while (doc != null) {

				//get sort field and doc uid and add to docstobesorted map
				//if docstobesorted includes key
				//get docid value and append it => so that sorting can be done
				var field = 'N/A'
				var docId = doc.getUniversalID()
				switch (sortField) {
					case "DATE": {
						var deliveredDate = getItemValue(doc, 'DeliveredDate') as DateTime
						if (deliveredDate != null) {
							field = deliveredDate.toJavaDate.time.toString

						}
					}
					case "SUBJECT": {
						var subject = doc.getItemValueString("Subject")
						if (subject != null && subject.length > 0) {
							field = subject
						}
					}
					case "FROM": {
						var fromAddr = doc.getItemValueString("From")
						Logger::log("FROM: " + fromAddr)
						if (fromAddr != null && fromAddr.length > 0) {
							field = fromAddr
						}
					}
				}
				if (docsToBeSorted.containsKey(field)) {
					var old = docsToBeSorted.get(field)
					docsToBeSorted.put(field, old + " " + docId)
				} else {
					docsToBeSorted.put(field, docId)
				}

				n = n + 1
				doc.recycle()
				doc = docCollection.getNthDocument(n)
			} //end while doc !=null

			var sortedDocsFields = sortMap(docsToBeSorted)

			if (sortOrder.equals("DESC")) {
				sortedDocsFields.reverse
			}

			for (_field : sortedDocsFields) {
				var id = docsToBeSorted.get(_field)
				if (id.contains(" ")) {
					var idList = id.split(" ")
					for (_id : idList) {
						newDocCollection.addDocument(database.getDocumentByUNID(_id))

					}
				} else {
					newDocCollection.addDocument(database.getDocumentByUNID(id))
				}
			}

			return newDocCollection

		} //end if sortfield

		return docCollection

	}

	def static DocumentCollection getMailDocCollection(DocumentCollection docCollection, Database database, Command cmd) {

		var listStartIndex = -1
		var listEndIndex = 10000
		var hasRangeIndex = false
		if (cmd.hasOption("start")) {
			listStartIndex = Integer::valueOf(cmd.getOptionValue("start"))
			hasRangeIndex = true
		}
		if (cmd.hasOption("end")) {
			listEndIndex = Integer::valueOf(cmd.getOptionValue("end"))
			hasRangeIndex = true
		}
		if (!hasRangeIndex) {
			return docCollection
		}

		var index = 1
		var newDocCollection = database.getProfileDocCollection("PROFILE_THAT_SHOULD_STILL_NOT_EXIST")
		var doc = docCollection.getNthDocument(index)
		while (doc != null) {
			if (index >= listStartIndex && index <= listEndIndex) {
				newDocCollection.addDocument(doc)

			}
			index = index + 1
			doc = docCollection.getNthDocument(index)
		}

		return newDocCollection
	}

	def static sortMap(Map<?, ?> map) {
		var keySet = map.keySet

		//var String[] vr = newArrayList
		var vr = newArrayList
		for (key : keySet) {
			vr.add(key as String)
		}
		vr.sort
	}

	def static getMailBody(Document mailDocument) {
		var body = getMailBody(mailDocument, "Body")
		if (body.nullOrEmpty) {
			body = getMailBody(mailDocument, "$Body")
		}
		body = removeMultipleEmptyLines(body)
		body
	}

	def static getMailBody(Document mailDocument, String fieldName) {

		var bodyItems = mailDocument.getItemValue(fieldName)

		//var body = mailDocument.getItemValueString("Body");
		var body = concatToString(bodyItems)
		if (body == "") {
			var item = mailDocument.getFirstItem(fieldName)
			if (item != null) {
				Logger::log("Getting body from: " + item.class.name)
			}
			switch (item) {
				case item instanceof RichTextItem: {
					var RichTextItem rtitem = item as RichTextItem
					body = rtitem.text
				}
				case item instanceof Item: {

					//likely a MIME email
					var Item _item = item as Item
					body = _item.getText()
					if (body == null) {
						var mimeEntity = _item.MIMEEntity
						body = getText(mimeEntity)
					}
				}
				case null: {
					body = "[Unable to get mail body]"
				}
			}
		}
		body
	}

	def static concatToString(Vector<?> items) {
		var txt = new StringBuffer()
		for (item : items) {
			txt.append(item.toString)
			txt.append(Configuration::LINE_SEPARATOR)
		}
		txt.toString
	}

	def static importance(Document mailDocument) {
		var importance = mailDocument.getItemValueString("Importance")
		if(importance == "1") return "High"
		if(importance == "2") return "Normal"
		null
	}

	def static String getAttachmentNames(Document doc, Session session, boolean replaceFile) {
		return doc.getAttachmentNames(null, null, false, session, replaceFile)
	}

	def static String getAttachmentNames(Document doc, String detachDir, String detachAttachmentName, boolean detachAll,
		Session session, boolean replaceFile) {

		Logger::log("Getting attachment names....")
		Logger::log("Document subject: " + getItemValue(doc, "Subject"))
		Logger::log("Document form: " + getItemValue(doc, "Form"))

		if ((detachAll || detachAttachmentName != null)) {
			println("Saving attachments from...")
			println("   " + getFormattedMail(doc, "!dtfs", 0, " ", session))
		}
		Logger::log("getAttachmentNames: " + doc)

		var hasNativeMIME = doc.hasItem("$NoteHasNativeMIME")

		var hasMIMEAttachments = hasNativeMIME && doc.hasEmbedded
		if (hasMIMEAttachments && (detachAll || detachAttachmentName != null)) {

			//detach native MIME attachments
			var names = ""
			names = detachMIMEAttachments(doc, session, detachDir, false, detachAttachmentName, replaceFile)
			return names
		}

		if (hasNativeMIME) {

			//for return only attachment names from native MIME mail
			return detachMIMEAttachments(doc, session, detachDir, true, detachAttachmentName, replaceFile)
		}

		if (hasNativeMIME) {
			var mimeEntity = doc.MIMEEntity
			Logger::log("Mime entity: " + mimeEntity)
		}

		var Item body = doc.getFirstItem("Body")
		if (body == null) {
			body = doc.getFirstItem("$Body");
		}
		var docAttachments = new Vector
		var str = new StringBuffer
		var someError = false

		if (body == null) {
			Logger::log("Body is null")
			Logger::log("  Subject: " + doc.getFirstItem("Subject"))
			Logger::log("  UNID   : " + doc.universalID)

		} else {
			Logger::log("Body item class: " + body.class.name)
			try {
				docAttachments = doc.embeddedObjects
				Logger::log("Attachments size from doc: " + docAttachments.size)
			} catch (Exception e) {
				someError = true
				str.append("[ERROR] Exception when getting attachments")

			}

			if (docAttachments.size == 0) {

				//cast item to RichTextItem and try to get attachments from that
				try {
					var RichTextItem rtitem = body as RichTextItem
					docAttachments = rtitem.embeddedObjects
					Logger::log("Attachments size from RichTextItem: " + docAttachments.size)
				} catch (Exception e) {

					//catch potential ClassCastException
					//reported bug in OpenNTF
					//http://www.openntf.org/internal/home.nsf/defect.xsp?action=openDocument&documentId=1152722592A1F85086257C3700354B19
					Logger::log(e)
				}
			}

			if (docAttachments.size == 0) {
				Logger::log("Check attachments from all items..")

				//still not found
				//check all items for attachments just in case
				var items = doc.items
				for (_item : items) {
					var item = _item as Item

					//Logger::log("item: " + item.name)
					if (item instanceof RichTextItem) {
						var rtitem = item as RichTextItem
						var eObjects = rtitem.embeddedObjects
						for (eo : eObjects) {
							docAttachments.add(eo)
						}
					} else {
						if (item.type == Item.ATTACHMENT) {

							//TODO: add option to disable rich text rendering
							//by using MIME
							Logger::log("File attachment found")
							Logger::log("  Doc has embedded: " + doc.hasEmbedded)
							var eObjects = doc.embeddedObjects
							for (eo : eObjects) {
								docAttachments.add(eo)
							}
							Logger::log("  Tota embedded: " + eObjects.size)
							Logger::log("  class: " + item.class.name)
							Logger::log("  name: " + item.name)
							Logger::log("  text: " + item.text)
							Logger::log("  value length: " + item.valueLength)
							Logger::log("  valuestring: " + item.valueString.trim)

						//add name to list
						//and extract if needed
						}
					}
				}
			}
		}

		for (o : docAttachments) {
			var EmbeddedObject a = o as EmbeddedObject
			Logger::log("EmbeddedObject: " + a.name)
			if (a.type == EmbeddedObject::EMBED_ATTACHMENT) {
				var fileName = a.name
				str.append(fileName)
				str.append(", ")
				var dir = "."
				if (detachAll || detachAttachmentName != null) {
					if (detachDir != null) {
						Logger::log("detach-dir: " + dir)
						dir = detachDir.replace('\\', '/')

						//dir=dir.replace('/','\\\\')
						if(!dir.endsWith("/")) dir = dir + "/"
						Logger::log("detach-dir modified: " + dir)
					} else {
						dir = "./"

					}

					//print "detachAll",detachAll
					var extractedFile = dir + fileName
					if (!replaceFile) { //if --replace option not specified, modify file name
						extractedFile = checkIfFileExists(extractedFile)
					}

					if (detachAll || fileName == detachAttachmentName) {
						a.extractFile(extractedFile)
						println("   Attachment saved: " + extractedFile + ".")
					}
				}
			}
		}

		var names = str.toString
		if (names.endsWith(",")) {
			names = names.substring(0, names.length - 1)
		}

		if (names.length == 0 && hasMIMEAttachments) {
			names = "[MIME attachments not yet implemented]"
		}
		return names
	}

	//(numberOfMails,outputTxt,mailDoc)=getMails(docCollection,opts,cmdOptions=cmdOpts,docReadIndex=readIndex)  
	def static List<?> getMails(DocumentCollection documentCollection, Command cmd, int docReadIndex, boolean noRange,
		Session session) {
		var outputFormatString = CommandLineArguments::getValue("output-format")
		var delimiter = CommandLineArguments::getValue("delim")
		if (CommandLineArguments::hasOption("tab")) {
			delimiter = "\t"
		}

		Logger::log("Doc read index: " + docReadIndex)
		var Document mailDocToBeRead = null

		/*
		if (docReadIndex > 0) {
			mailDocToBeRead = documentCollection.getNthDocument(docReadIndex-1)
			return #[0, "", mailDocToBeRead]
		}
*/
		var doc = documentCollection.getFirstDocument();
		var index = 0
		var listStartIndex = -1
		var listEndIndex = 10000
		var hasRangeIndex = false
		if (!noRange) {

			if (cmd.hasOption("start")) {
				listStartIndex = Integer::valueOf(cmd.getOptionValue("start"))
				hasRangeIndex = true
			}
			if (cmd.hasOption("end")) {
				listEndIndex = Integer::valueOf(cmd.getOptionValue("end"))
				hasRangeIndex = true
			}
		}

		var outputTxt = new StringBuffer
		var isReceivedMail = false
		var processDoc = false
		while (doc != null) {
			isReceivedMail = true
			if (docReadIndex == index) {
				mailDocToBeRead = doc

			}
			processDoc = false

			var deliveredDate = DominoUtils::getItemValue(doc, "DeliveredDate")
			if(deliveredDate == null) isReceivedMail = false
			if (isReceivedMail) {
				processDoc = true
				if (hasRangeIndex) {
					if (index >= listStartIndex && index <= listEndIndex) {
						processDoc = true
					} else {
						processDoc = false
					}
				}

				//	}
				Logger::log(
					"index: " + index + " Prosessdoc: " + processDoc + ", start: " + listStartIndex + ", end: " +
						listEndIndex)
				if (processDoc) {
					if (outputFormatString != null) {
						outputTxt.append(getFormattedMail(doc, outputFormatString, index, delimiter, session))

					} else {
						outputTxt.append("[" + index + "] ")
						outputTxt.append(deliveredDate)
						outputTxt.append(Configuration::LINE_SEPARATOR)
						outputTxt.append("  ")
						if (doc.isSentByAgent()) {

							//fromField=doc.getItemValueString("Principal")
							outputTxt.append("From (Sent by agent?, probable sender): ")

						} else {

							//fromField=doc.getItemValueString("From")
							//principalField=doc.getItemValueString("Principal")
							//if principalField is not None and not "":
							//  if fromField!=principalField:
							//    principalField=principalField.strip()
							//    if principalField!=(""):
							//      fromField=principalField+" ("+fromField+")"
							outputTxt.append("From: ")

						}

						var fromField = doc.getItemValueString("From")
						var principalField = doc.getItemValueString("Principal")
						if (principalField != null && principalField != "") {
							if (fromField != principalField) {
								principalField = principalField.trim()
								if (principalField != "") {
									fromField = principalField + " (" + fromField + ")"
								}
							}
						}

						outputTxt.append(fromField)
						outputTxt.append(Configuration::LINE_SEPARATOR)
						outputTxt.append("  ")

						var subject = doc.getItemValueString("Subject")
						outputTxt.append("Subject: ")
						outputTxt.append(subject)
						outputTxt.append(Configuration::LINE_SEPARATOR)
						outputTxt.append("  ")

						var importance = importance(doc)

						if (importance == "High") {
							outputTxt.append("High Importance")
							outputTxt.append(Configuration::LINE_SEPARATOR)
							outputTxt.append("  ")

						}

						outputTxt.append("Attachments: ")
						outputTxt.append(getAttachmentNames(doc, session, false))
						outputTxt.append(Configuration::LINE_SEPARATOR)
						outputTxt.append("  ")
					}
					outputTxt.append(Configuration::LINE_SEPARATOR)

				//index = index + 1
				}
				index = index + 1

			}
			doc = documentCollection.getNextDocument()

		}

		var mailTxt = outputTxt.toString
		#[index, mailTxt, mailDocToBeRead]

	//return rv//(index,mailTxt,mailDocToBeRead)
	}

	def static getFormattedMail(Document mailDoc, String _formatString, int index, String _delimiter, Session session) {
		var delimiter = _delimiter
		if(delimiter == null) delimiter = ";"
		var txt = new StringBuffer
		if (_formatString.indexOf("!") == -1) {
			txt.append(index)
			txt.append(delimiter)
		}
		var datetime = DominoUtils.getItemValue(mailDoc, "DeliveredDate") as DateTime

		//mailDoc.getItemValueDateTimeArray("DeliveredDate").get(0) as DateTime
		var formatString = _formatString.toCharArray
		for (_chr : formatString) {
			var chr = Character::toString(_chr)
			switch (chr) {
				case 'd': {
					if (datetime == null) {
						txt.append("N/A")
					} else {
						txt.append(datetime.getDateOnly())
					}
					txt.append(delimiter)

				}
				case 't': {
					if (datetime == null) {
						txt.append("N/A")
					} else {
						txt.append(datetime.getTimeOnly())
					}
					txt.append(delimiter)

				}
				case 'Z': {

					if (datetime == null) {
						txt.append("N/A")
					} else {
						txt.append(datetime.getTimeZone())
					}

					//if datetime.isDST():
					//  txt.write(' (DST)')
					txt.append(delimiter)
				}
				case 'g': {

					if (datetime == null) {
						txt.append("N/A")
					} else {
						txt.append(datetime.getGMTTime())
					}

					txt.append(delimiter)
				}
				case 's': {

					var value = mailDoc.getItemValueString("Subject")
					txt.append(value)
					txt.append(delimiter)
				}
				case 'f': {
					var value = mailDoc.getItemValueString("From")
					txt.append(value)
					txt.append(delimiter)
				}
				case 'I': {
					var value = mailDoc.getItemValueString("INetFrom")
					txt.append(value)
					txt.append(delimiter)

				}
				case 'm': {
					var value = mailDoc.getItemValueString("$Mailer")
					txt.append(value)
					txt.append(delimiter)

				}
				case 'D': {
					var value = mailDoc.getItemValueString("$MessageID")
					txt.append(value)
					txt.append(delimiter)

				}
				case 'S': {
					var value = mailDoc.getSize()
					txt.append(value)
					txt.append(delimiter)

				}
				case 'a': {
					var value = getAttachmentNames(mailDoc, session, false) as String
					if (value.nullOrEmpty) {
						value = "n/a"
					}
					txt.append(value)
					txt.append(delimiter)

				}
			}
		}

		var rv = txt.toString
		return rv.substring(0, rv.length - delimiter.length) //#removes delimiter at the end
	}

	def static Document getLatestMailDocument(DocumentCollection documentCollection) {
		var doc = documentCollection.getLastDocument()
		while (doc != null) {
			var deliveredDate = getItemValue(doc, "DeliveredDate")
			if (deliveredDate != null) {
				return doc
			}
			doc = documentCollection.getPrevDocument()

		}
		null //should never happen

	}

	def static Vector<String> readRecipientFile(String recipientFile) {

		var f = new File(recipientFile)
		var fr = new BufferedReader(new FileReader(f))

		var recipients = new Vector()

		var line = fr.readLine
		while (line != null) {
			recipients.add(line)
			line = fr.readLine
		}
		fr.close()

		var recipientStr = recipients.join(",")
		var _recipients = recipientStr.split(',')
		var recipientVector = new Vector<String>()
		for (recipient : _recipients) {
			recipientVector.add(recipient.trim())

		}
		recipientVector
	}

	def static getText(MIMEEntity mime) {

		var text = new StringBuffer()

		// If multipart MIME entity
		if (mime.getContentType().equals("multipart")) {

			var MIMEEntity child1 = mime.getFirstChildEntity();
			var textAlreadyExtracted = false
			while (child1 != null) {

				var contentType = child1.contentType
				if (contentType == "text" && !textAlreadyExtracted) {
					if (CommandLineArguments::hasOption("all-mime-texts")) {
						text.append("==MIME part: " + contentType + "/" + child1.contentSubType)
						text.append(Configuration::LINE_SEPARATOR)
					}

					text.append(getContentAsText(child1))
					if (!CommandLineArguments::hasOption("all-mime-texts")) {
						textAlreadyExtracted = true
					}
				}

				var MIMEEntity child2 = child1.getFirstChildEntity();

				if (child2 == null) {
					child2 = child1.getNextSibling();
					if (child2 == null) {
						child2 = child1.getParentEntity();
						if (child2 != null)
							child2 = child2.getNextSibling();
					}
				}
				child1 = child2;
			}
		} else {
			var contentType = mime.contentType
			if (contentType == "text") {
				text.append(getContentAsText(mime))
			}
		}

		text.toString
	}

	def static getContentAsText(MIMEEntity mime) {
		var text = new StringBuffer()
		var subContentType = mime.contentSubType
		Logger::log(subContentType)
		switch (subContentType) {
			case "html": {
				var txt = mime.contentAsText
				if (!CommandLineArguments::hasOption("no-striphtmltags")) {

					//remove html tags
					//check if HTML has HEAD tag
					//if so, remove
					var lowerTxt = txt.toLowerCase
					if (lowerTxt.contains("<head>")) {
						var start = lowerTxt.indexOf("<head>")
						var end = lowerTxt.indexOf("</head>")
						txt = txt.substring(0, start) + txt.substring(end, txt.length)
					}

					txt = txt.replaceAll("<[^>]*>", "")
				}
				text.append(txt)
			}
			default: {
				text.append(mime.contentAsText)
			}
		}
		text.append(Configuration::LINE_SEPARATOR)
		text.toString
	}

	def static removeMultipleEmptyLines(String txt) {
		var text = new StringBuffer
		var emptyLastLine = false
		var txtLines = txt.split(Configuration::LINE_SEPARATOR)
		for (_line : txtLines) {
			var line = _line.trim
			if (line.nullOrEmpty) {
				if (!emptyLastLine) {
					text.append(Configuration::LINE_SEPARATOR)
				}
				emptyLastLine = true
			} else {
				emptyLastLine = false
				text.append(line)
				text.append(Configuration::LINE_SEPARATOR)
			}
		}
		text.toString
	}

	def static detachMIMEAttachments(Document doc, Session session, String _detachDir, boolean getNamesOnly,
		String attachmentName, boolean replaceFile) {
		var names = ""
		session.setConvertMIME(false)

		var MIMEEntity mime = doc.getMIMEEntity();

		if (mime != null) {

			// If multipart MIME entity
			if (mime.getContentType().equals("multipart")) {

				var MIMEEntity child1 = mime.getFirstChildEntity();

				while (child1 != null) {

					var contentType = child1.contentType
					if (contentType == "application") {
						var String name = null
						name = detachAttachment(session, child1, _detachDir, getNamesOnly, attachmentName, replaceFile)

						if (!name.nullOrEmpty) {
							names = names + name + ","
						}
					}

					var MIMEEntity child2 = child1.getFirstChildEntity();

					if (child2 == null) {

						child2 = child1.getNextSibling();
						if (child2 == null) {
							child2 = child1.getParentEntity();
							if (child2 != null)
								child2 = child2.getNextSibling();
						}
					}
					child1 = child2;
				}
			} else {
				var contentType = mime.contentType
				if (contentType == "application") {
					var String name = null
					name = detachAttachment(session, mime, _detachDir, getNamesOnly, attachmentName, replaceFile)
				}
			}
		}

		session.setConvertMIME(true)
		if (names != null && names.endsWith(",")) {
			names = names.substring(0, names.length - 1)
		}
		names
	}

	def static detachAttachment(Session session, MIMEEntity mimeEntity, String _detachDir, boolean getNamesOnly,
		String attachmentName, boolean replaceFile) {

		var String name = null
		var detachDir = _detachDir
		if (detachDir == null) {
			detachDir = "./"
		}
		var _dir = new File(detachDir)
		var dir = _dir.canonicalPath
		if (!dir.endsWith("/") && !dir.endsWith("\\")) {
			dir = dir + "/"
		}

		mimeEntity.decodeContent
		printMIMEHeaders(mimeEntity)
		var header = mimeEntity.getNthHeader("Content-Disposition")
		if (header != null) {
			if (header.headerVal == "attachment") {

				//attachment exists
				//note: inline attachments not supported
				name = header.getParamVal("filename")
				name = name.replace("\"", "").trim

				//save attachment only if saving all attachments or attachment name
				//contains given attachmentname
				if (!getNamesOnly && (attachmentName == null || name.contains(attachmentName))) {
					var Stream stream = session.createStream();

					var filePath = dir + name
					if (!replaceFile) { //if --replace option not specified, modify file name
						filePath = checkIfFileExists(filePath)
					}
					if (stream.open(filePath, "binary")) {
						mimeEntity.getContentAsBytes(stream)
						stream.close
						println("   Attachment saved: " + name + ".")
					} else {
						println("   Failed to detach: " + name + ".")
					}
				}
			}
		}
		name
	}

	def static printMIMEHeaders(MIMEEntity mimeEntity) {
		var headers = mimeEntity.headerObjects
		Logger::log("=== MIME Entity: " + mimeEntity.toString)
		for (_header : headers) {
			var MIMEHeader header = _header as MIMEHeader

			Logger::log(header.headerName)
			Logger::log("  value : " + header.getHeaderVal())
			Logger::log("  value and params: " + header.headerValAndParams)
		}
	}

	def static checkIfFileExists(String filePath) {
		var newFilePath = filePath
		var seq = 2
		var f = new File(newFilePath)
		while (f.exists) {

			//modify file
			var index = filePath.lastIndexOf(".")
			if (index == -1) {

				//not found file type, just add sequence to end
				newFilePath = String.format("%s_(%d)", filePath, seq)
			} else {

				//found fule type, add seq before type
				newFilePath = String.format("%s_(%d)%s", filePath.substring(0, index), seq, filePath.substring(index))
			}
			seq = seq + 1
			f = new File(newFilePath)
		}
		return newFilePath

	}

	def static printFields(Document document) {

		//list fields and types in the document
		var items = document.items

		var columnFields = newHashMap()
		for (_item : items) {
			var item = _item as Item
			var name = item.name
			var _type = item.type
			var className = item.class.name
			var type = ""
			switch (_type) {
				case Item.ACTIONCD: {
					type = "ACTIONCD"
				}
				case Item.ASSISTANTINFO: {
					type = "ASSISTANTINFO"
				}
				case Item.ATTACHMENT: {
					type = "ATTACHMENT (file attachment)"
				}
				case Item.AUTHORS: {
					type = "AUTHORS"
				}
				case Item.COLLATION: {
					type = "COLLATION"
				}
				case Item.DATETIMES: {
					type = "DATETIMES (date-time or range of date-time values)"
				}
				case Item.EMBEDDEDOBJECT: {
					type = "EMBEDDEDOBJECT"
				}
				case Item.ERRORITEM: {
					type = "ERRORITEM (error occurred while getting type)"
				}
				case Item.FORMULA: {
					type = "FORMULA (Domino formula)"
				}
				case Item.HTML: {
					type = "HTML (HTML source text)"
				}
				case Item.ICON: {
					type = "ICON"
				}
				case Item.LSOBJECT: {
					type = "LSOBJECT"
				}
				case Item.MIME_PART: {
					type = "MIME_PART"
				}
				case Item.NAMES: {
					type = "NAMES"
				}
				case Item.NOTELINKS: {
					type = "NOTELINKS (link to a database, view, or document)"
				}
				case Item.NOTEREFS: {
					type = "NOTEREFS (reference to the parent document)"
				}
				case Item.NUMBERS: {
					type = "NUMBERS (number or number list)"
				}
				case Item.OTHEROBJECT: {
					type = "OTHEROBJECT"
				}
				case Item.QUERYCD: {
					type = "QUERYCD"
				}
				case Item.READERS: {
					type = "READERS"
				}
				case Item.RFC822TEXT: {
					type = "RFC822TEXT"
				}
				case Item.RICHTEXT: {
					type = "RICHTEXT"

				}
				case Item.SIGNATURE: {
					type = "SIGNATURE"
				}
				case Item.TEXT: {
					type = "TEXT (text or text list)"
				}
				case Item.UNAVAILABLE: {
					type = "UNAVAILABLE"
				}
				case Item.UNKNOWN: {
					type = "UNKNOWN"
				}
				case Item.USERDATA: {
					type = "USERDATA"
				}
				case Item.USERID: {
					type = "USERID"
				}
				case Item.VIEWMAPDATA: {
					type = "VIEWMAPDATA"
				}
				case Item.VIEWMAPLAYOUT: {
					type = "VIEWMAPLAYOUT"
				}
				default: {
					type = "undetermined"
				}
			}
			columnFields.put(name, type + "," + className)

		}

		var sortedNames = columnFields.keySet.sort
		var index = 1
		var printIndex = false
		println("Total fields: " + sortedNames.size)
		var list = newArrayList()
		list.add("NAME")
		list.add("TYPE")

		//list.add("Java class")
		Output.prettyPrintln(list, 30, 0)
		for (name : sortedNames) {
			list = newArrayList()
			var indexS = ""
			if (printIndex) {
				if (index < 10) {
					indexS = index + " : "
				} else {
					indexS = index + ": "
				}
			}
			list.add(indexS + name)
			var typeAndClass = columnFields.get(name).split(",")
			list.add(typeAndClass.get(0))

			//list.add(typeAndClass.get(1))
			Output.prettyPrintln(list, 30, 0)
			index = index + 1
		}
	}

	def static printFieldValue(Document document, String fieldName) {
		if (!document.hasItem(fieldName)) {
			println("Field " + fieldName + " does not exist.")
			return
		}

		println("Field " + fieldName + ":")
		var itemValues = document.getItemValue(fieldName)

		if (itemValues.size == 0) {
			println("  N/A")
		}
		for (value : itemValues) {
			println("  " + value)
		}
	}

}
